package gregapi.block.multitileentity.example;

import static gregapi.data.CS.NBT_OWNER;
import static gregapi.data.CS.T;
import static gregapi.data.CS.ZL_BYTE;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_AddToolTips;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_GetPlayerRelativeBlockHardness;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_OnPlaced;
import gregapi.block.multitileentity.MultiTileEntityContainer;
import gregapi.data.LH;
import gregapi.data.LH.Chat;
import gregapi.network.INetworkHandler;
import gregapi.network.IPacket;
import gregapi.old.Textures;
import gregapi.render.BlockTextureDefault;
import gregapi.render.BlockTextureMulti;
import gregapi.render.IIconContainer;
import gregapi.render.ITexture;
import gregapi.tileentity.base.TileEntityBase07Paintable;
import gregapi.util.UT;

import java.util.List;
import java.util.UUID;

import net.minecraft.block.Block;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.world.World;

/**
 * @author Gregorius Techneticies
 * 
 * An example implementation of a private Block.
 */
public class MultiTileEntityBunkerBlock extends TileEntityBase07Paintable implements IMTE_GetPlayerRelativeBlockHardness, IMTE_AddToolTips, IMTE_OnPlaced {
	public UUID mOwner;
	
    @Override
	public void readFromNBT2(NBTTagCompound aNBT) {
    	super.readFromNBT2(aNBT);
    	if (aNBT.hasKey(NBT_OWNER)) mOwner = UUID.fromString(aNBT.getString(NBT_OWNER));
    }
    
    @Override
	public void writeToNBT2(NBTTagCompound aNBT) {
    	super.writeToNBT2(aNBT);
        if (mOwner != null) aNBT.setString(NBT_OWNER, mOwner.toString());
    }
    
	@Override
	public void addToolTips(List aList, ItemStack aStack, boolean aF3_H) {
		aList.add(Chat.ORANGE + LH.get(LH.OWNER_CONTROLLED));
	}
	
	@Override
	public boolean allowInteraction(Entity aEntity) {
		if (aEntity != null) {
			UUID tUUID = aEntity.getUniqueID();
			return mOwner == null || (tUUID != null && tUUID.equals(mOwner));
		}
		return mOwner == null;
	}
	
    @Override
	public boolean onPlaced(ItemStack aStack, EntityPlayer aPlayer, MultiTileEntityContainer aMTEContainer, World aWorld, int aX, int aY, int aZ, byte aSide, float aHitX, float aHitY, float aHitZ) {
		if (aPlayer != null) mOwner = aPlayer.getUniqueID();
		return T;
    }
	
	@Override
	public IPacket getClientDataPacket(boolean aSendAll) {
		if (aSendAll) {
			byte[] tName = mOwner==null?ZL_BYTE:mOwner.toString().getBytes(), tData = new byte[tName.length + 3];
			tData[0] = (byte)UT.Code.getR(mRGBa); tData[1] = (byte)UT.Code.getG(mRGBa); tData[2] = (byte)UT.Code.getB(mRGBa);
			for (short i = 0; i < tName.length; i++) tData[i+3] = tName[i];
			return getClientDataPacketByteArray(aSendAll, tData);
		}
		return null;
	}
	
	@Override
	public boolean receiveDataByteArray(byte[] aData, INetworkHandler aNetworkHandler) {
		mRGBa = UT.Code.getRGBInt(new short[] {UT.Code.unsignB(aData[0]), UT.Code.unsignB(aData[1]), UT.Code.unsignB(aData[2])});
		if (aData.length > 3) mOwner = UUID.fromString(new String(aData, 4, aData.length-4));
		return T;
	}
	
	public static IIconContainer sColored = new Textures.BlockIcons.CustomIcon("bunker/block/colored"), sOverlay = new Textures.BlockIcons.CustomIcon("bunker/block/overlay");
	
	@Override public ITexture getTexture2(Block aBlock, int aRenderPass, byte aSide) {return BlockTextureMulti.get(BlockTextureDefault.get(sColored, mRGBa), BlockTextureDefault.get(sOverlay));}
	@Override public float getPlayerRelativeBlockHardness(EntityPlayer aPlayer, float aOriginal) {return allowInteraction(aPlayer) ? Math.max(aOriginal, 0.0001F) : 0.0F;}
	@Override public String getTileEntityName() {return "gt.multitileentity.bunker.block";}
}